import json
from dataclasses import dataclass
from typing import Optional, List

import typer
from rich.console import Console
from rich.table import Table

from amora.compilation import amora_model_for_path
from amora.config import settings
from amora.models import (
    list_model_files,
    AmoraModel,
)
from amora.providers.bigquery import dry_run, DryRunResult


app = typer.Typer()


@dataclass
class ResultItem:
    model: AmoraModel
    dry_run_result: Optional[DryRunResult] = None

    def as_dict(self):
        return {
            "depends_on": self.depends_on,
            "has_source": self.has_source,
            "materialization_type": self.materialization_type,
            "model_name": self.model_name,
            "referenced_tables": self.referenced_tables,
            "total_bytes": self.total_bytes,
        }

    @property
    def model_name(self):
        return self.model.__name__

    @property
    def has_source(self):
        return self.model.source() is not None

    @property
    def depends_on(self) -> List[str]:
        return sorted(
            [dependency.__name__ for dependency in self.model.dependencies()]
        )

    @property
    def total_bytes(self) -> Optional[int]:
        if self.dry_run_result:
            return self.dry_run_result.total_bytes
        return None

    @property
    def referenced_tables(self) -> List[str]:
        if self.dry_run_result:
            return self.dry_run_result.referenced_tables
        return []

    @property
    def materialization_type(self) -> Optional[str]:
        if self.has_source:
            return self.model.__model_config__.materialized.value
        else:
            return None


@app.command(name="list")
def list_models(
    format: str = typer.Option(
        "table",
        help="Output format. Options: json,table",
    ),
    with_total_bytes: bool = typer.Option(
        False,
        help="Uses BigQuery query dry run feature "
        "to gather model total bytes processed information",
    ),
) -> None:
    """
    List the resources in your project

    """
    results = []
    for model_file_path in list_model_files():
        model = amora_model_for_path(model_file_path)
        if with_total_bytes:
            result_item = ResultItem(model=model, dry_run_result=dry_run(model))
        else:
            result_item = ResultItem(model=model, dry_run_result=None)

        results.append(result_item)

    if format == "table":
        table = Table(
            show_header=True,
            header_style="bold",
            show_lines=True,
            width=settings.CLI_CONSOLE_MAX_WIDTH,
            row_styles=["none", "dim"],
        )

        table.add_column("Model name", style="green bold", no_wrap=True)
        table.add_column("Total bytes", no_wrap=True)
        table.add_column("Referenced tables", no_wrap=True)
        table.add_column("Depends on", no_wrap=True)
        table.add_column("Has source?", no_wrap=True, justify="center")
        table.add_column("Materialization", no_wrap=True)

        for result in results:
            table.add_row(
                result.model_name,
                f"{result.total_bytes or '-'}",
                " , ".join(result.referenced_tables) or "-",
                " , ".join(result.depends_on) or "-",
                "🟢" if result.has_source else "🔴",
                result.materialization_type or "-",
            )

        console = Console(width=settings.CLI_CONSOLE_MAX_WIDTH)
        console.print(table)

    elif format == "json":
        output = {"models": [result.as_dict() for result in results]}
        typer.echo(json.dumps(output))


def main():
    return app()



import typer

from amora.cli import compile, list, materialize, model, test

app = typer.Typer(
    help="Amora Data Build Tool enables engineers to transform data in their warehouses "
    "by defining schemas and writing select statements with SQLAlchemy. Amora handles turning these "
    "select statements into tables and views"
)

app.add_typer(compile.app, name="compile")
app.add_typer(list.app, name="list")
app.add_typer(materialize.app, name="materialize")
# app.add_typer(model.app)
app.add_typer(test.app, name="test")


if __name__ == "__main__":
    app()



from dataclasses import dataclass
from pathlib import Path
from typing import List, Any, Iterable, Optional

from google.cloud.bigquery import Table, Client, QueryJobConfig
import matplotlib.pyplot as plt
import networkx as nx

from amora.compilation import amora_model_for_target_path
from amora.config import settings
from amora.models import list_target_files, AmoraModel, MaterializationTypes


@dataclass
class Task:
    sql_stmt: str
    model: AmoraModel
    target_file_path: Path

    @classmethod
    def for_target(cls, target_file_path: Path) -> "Task":
        return cls(
            sql_stmt=target_file_path.read_text(),
            model=amora_model_for_target_path(target_file_path),
            target_file_path=target_file_path,
        )

    def __repr__(self):
        return f"{self.model.__name__} -> {self.sql_stmt}"


class DependencyDAG(nx.DiGraph):
    def __iter__(self):
        # todo: validar se podemos substituir por graphlib
        return nx.topological_sort(self)

    @classmethod
    def from_tasks(cls, tasks: Iterable[Task]) -> "DependencyDAG":
        dag = cls()

        for task in tasks:
            dag.add_node(task.model.__name__)
            for dependency in getattr(task.model, "__depends_on__", []):
                dag.add_edge(dependency.__name__, task.model.__name__)

        return dag

    def draw(self) -> None:
        nx.draw_spectral(
            self,
            with_labels=True,
            font_weight="bold",
            font_size="12",
            linewidths=4,
            node_size=150,
            node_color="white",
            font_color="green",
        )
        plt.show()


def materialize(sql: str, model: AmoraModel) -> Optional[Table]:
    materialization = model.__model_config__.materialized
    table_id = f"{settings.TARGET_PROJECT}.{settings.TARGET_SCHEMA}.{model.__tablename__}"

    if materialization == MaterializationTypes.view:
        view = Table(table_id)
        view.view_query = sql

        return Client().create_table(view, exists_ok=True)
    elif materialization == MaterializationTypes.table:
        client = Client()
        query_job = client.query(
            sql,
            job_config=QueryJobConfig(
                destination=table_id, write_disposition="WRITE_TRUNCATE"
            ),
        )

        query_job.result()
        return client.get_table(table_id)
    elif materialization == MaterializationTypes.ephemeral:
        return None
    else:
        raise ValueError(
            f"Invalid model materialization configuration. "
            f"Valid types are: `{', '.join((m.name for m in MaterializationTypes))}`. "
            f"Got: `{materialization}`"
        )



from dataclasses import dataclass
from enum import Enum, auto
from inspect import getfile
from pathlib import Path
from typing import Iterable, List, Optional, Type, Union

from amora.config import settings
from dbt.adapters.bigquery.impl import PartitionConfig
from sqlalchemy import MetaData
from sqlmodel import SQLModel, Field, select
from sqlmodel.sql.expression import Select, SelectOfScalar

Compilable = Union[Select, SelectOfScalar]
PartitionConfig = PartitionConfig

select = select


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class MaterializationTypes(AutoName):
    ephemeral = auto()
    view = auto()
    table = auto()


@dataclass
class ModelConfig:
    materialized: MaterializationTypes = MaterializationTypes.view
    partition_by: Optional[PartitionConfig] = None
    cluster_by: Optional[Union[str, List[str]]] = None
    tags: Optional[List[str]] = None


def list_files(path: Union[str, Path], suffix: str) -> Iterable[Path]:
    yield from Path(path).rglob(f"*{suffix}")


def list_model_files() -> Iterable[Path]:
    return list_files(settings.MODELS_PATH, suffix=".py")


def list_target_files() -> Iterable[Path]:
    return list_files(settings.TARGET_PATH, suffix=".sql")


# todo: Como defino "Objeto que tenha 2 atributos: `source` e `output`" ?
def is_py_model(obj) -> bool:
    return hasattr(obj, "source") and hasattr(obj, "output")


# registry = registry()

metadata = MetaData(
    schema=f"{settings.TARGET_PROJECT}.{settings.TARGET_SCHEMA}"
)


class AmoraModel(SQLModel):
    __depends_on__: List["AmoraModel"] = []
    __model_config__ = ModelConfig(materialized=MaterializationTypes.view)
    __table_args__ = {"extend_existing": True}
    metadata = metadata

    @classmethod
    def dependencies(cls) -> Iterable["AmoraModel"]:
        source = cls.source()
        if source is None:
            return []

        # todo: Remover necessidade de __depends_on__ inspecionando a query e chegando ao modelo de origem
        # tables: List[Table] = source.froms

        return cls.__depends_on__

    @classmethod
    def source(cls) -> Optional[Compilable]:
        """
        Called when `amora compile` is executed, Amora will build this model
        in your data warehouse by wrapping it in a `create view as` or `create table as` statement.

        Return `None` for defining models for tables/views that already exist on the data warehouse
        and shouldn't be managed by Amora.

        Returning a `Compilable`, which is a sqlalchemy select statement
        :return:
        """
        return None

    @classmethod
    def target_path(cls, model_file_path: Union[str, Path]) -> Path:
        # {settings.dbt_models_path}/a_model/a_model.py -> a_model/a_model.py
        strip_path = settings.MODELS_PATH
        relative_model_path = str(model_file_path).split(strip_path)[1][1:]
        # a_model/a_model.py -> ~/project/amora/target/a_model/a_model.sql
        target_file_path = Path(settings.TARGET_PATH).joinpath(
            relative_model_path.replace(".py", ".sql")
        )

        return target_file_path

    @classmethod
    def model_file_path(cls) -> Path:
        return Path(getfile(cls))



from typing import Optional, List

import typer

from amora.compilation import amora_model_for_path, compile_statement
from amora.models import (
    list_model_files,
    AmoraModel,
)


Models = List[str]
app = typer.Typer()


@app.command()
def compile(
    models: Optional[Models] = typer.Option(
        [],
        "--model",
        help="A model to be compiled. This option can be passed multiple times.",
    )
) -> None:
    """
    Generates executable SQL from model files. Compiled SQL files are written to the `./target` directory.
    """
    for model_file_path in list_model_files():
        if models and model_file_path.stem not in models:
            continue

        try:
            AmoraModel_class = amora_model_for_path(model_file_path)
        except ValueError:
            continue

        if not issubclass(AmoraModel_class, AmoraModel):  # type: ignore
            continue

        source_sql_statement = AmoraModel_class.source()
        if source_sql_statement is None:
            typer.echo(f"⏭ Skipping compilation of model `{model_file_path}`")
            continue

        target_file_path = AmoraModel_class.target_path(model_file_path)
        typer.echo(
            f"🏗 Compiling model `{model_file_path}` -> `{target_file_path}`"
        )

        content = compile_statement(source_sql_statement)
        target_file_path.write_text(content)


if __name__ == "__main__":
    app()
